import io
import logging
import sys
from typing import Dict

from agents_for_all.tools.base_tool import Tool


class Python(Tool):
    """
    A tool that can execute python codes.

    Accepts code via input JSON and executes it on the host system using python..
    """

    @property
    def name(self) -> str:
        """
        Python
        """
        return "Python"

    @property
    def description(self) -> str:
        """
        Explains that this tool executes raw Python code.
        """
        return (
            "A tool which can execute python commands commands on the host. "
            "It can generate values or produce side effects (e.g., creating files, etc). "
            'Provide the command to execute as: {"code": "<your python command>"}'
            'All the imports should be in the value for "code"'
            "Be careful of indentations and possible errors"
            "The only thing this tool can handle is actual code. The LLM shouldn't output anything else."
            "This means no explanation. This means no code inside ```python```. Just pure python code to execute."
        )

    def execute(self, input_json: Dict) -> str:
        """
        Execute the given code string using python.

        Args:
            input_json (Dict): Must contain a "code" key with the python command as value.

        Returns:
            str: The output or error string from the execution.
        """
        command = input_json.get("code")
        if not command:
            return "Error: 'code' key missing from input_json."

        old_stdout = sys.stdout
        sys.stdout = buffer = io.StringIO()

        try:
            exec(command)
        except Exception as e:
            logging.info(f"Execution failed with exception: {str(e)}")
            raise e
        finally:
            sys.stdout = old_stdout
        output = buffer.getvalue()
        return output



from abc import ABC, abstractmethod


class Model(ABC):
    """
    Abstract base class for models or model connectors to be exact.
    The models are initalized using its subclasses and can be used in agents.
    Some models are:
    - Direct (with api_endpoint, query_parameter_name, and other_parameters)
    - OpenAI
    - Antropic
    - Gemini
    """

    @abstractmethod
    def get_response(self, query: str) -> str:
        """
        Get response from the LLM based on the given query.

        Args:
            query (str): The query the LLM should respond to.

        Returns:
            str: The response to the query from the LLM.
        """
        raise NotImplementedError



import json
import logging
import re
from typing import Any, Dict, List

from agents_for_all.llms.base_model import Model
from agents_for_all.tools.base_tool import Tool

logging.basicConfig(level=logging.INFO)

DEFAULT_SUBDIVISION_PROMPT = """
You are a task-planning AI. Give a JSON array preferring tools over llm given the following user request:

{action}

And a list of available tools described below:

{tool_descriptions}

Break the task into steps. For each step, specify:
- "type": "tool" or "llm"
- "name": name of the tool (if type is tool)
- "input_json": dictionary input for the tool (if type is tool)
- "query": string to send to LLM (if type is llm)

Respond only with a JSON array like:
[
  {{"type": "tool", "name": "Computer", "input_json": {{"code": "echo hello"}}}},
  {{"type": "llm", "query": "Summarize what was just done."}}
]

Do not respond with anything other than the JSON array. No need for explanation or anything.
Do not explain the steps or say anything in the beginning. No need to tag the json with ```json.
Just give the JSON array as told.
"""

ERROR_CORRECTION_PROMPT = """
Last time, with these steps we got an error.

Steps:
{steps}

{error}
"""


class Agent:
    """
    Agent class which can 'do' actions using llm and tools.

    Example code:

    .. code-block:: python

        from agents_for_all import Agent
        from agents_for_all.llms.direct import Direct
        from agents_for_all.tools.python import Python

        llm = Direct(
            api_endpoint="http://127.0.0.1:1234/v1/chat/completions",
            model="deepseek-r1-distill-llama-8b"
        )
        python = Python()

        agent = Agent(llm=llm, tools=[python])
        print(agent.do("Create a file with system date as the filename and txt as extension."))

    """

    def __init__(self, llm: Model, tools: List[Tool], max_retries: int = 10):
        self.llm = llm
        self.tools = {tool.name: tool for tool in tools}
        self.max_retries = max_retries

    def _force_json_response(self, llm_response: str):
        try:
            return json.loads(llm_response)
        except json.JSONDecodeError:
            # Try to extract JSON if it was wrapped in text
            match = re.search(r"\[.*\]", llm_response, re.DOTALL)
            if match:
                return json.loads(match.group())
            raise

    def _subdivide_task(
        self, action: str, error: str = "", steps: Dict | None = None
    ) -> List[Dict[str, Any]]:
        tool_descriptions = "\n".join(
            [f"- {tool.name}: {tool.description}" for tool in self.tools.values()]
        )
        prompt = DEFAULT_SUBDIVISION_PROMPT.format(
            action=action, tool_descriptions=tool_descriptions, error=error
        )
        prompt += (
            ERROR_CORRECTION_PROMPT.format(steps=steps, error=error) if error else ""
        )
        response = self.llm.get_response(prompt)
        return self._force_json_response(response)

    def do(self, action: str) -> str:
        """
        Do the given action using tools available and the llm specified.

        Args:
            action (str): The action to perform.

        Returns:
            str: The output from doing the action or a brief summary of what happened by performing the action.
        """

        attempt = 0
        steps = []
        error = ""
        while attempt < self.max_retries:
            try:
                steps = self._subdivide_task(action=action, steps=steps, error=error)
                results = []
                history = []

                for idx, step in enumerate(steps):
                    step_type = step.get("type")

                    if step_type == "tool":
                        name = step.get("name")
                        input_json = step.get("input_json")
                        tool = self.tools.get(name)
                        if not tool:
                            message = f"Tool '{name}' not found."
                            logging.error(message)
                            results.append(message)
                            history.append(message)
                            continue

                        logging.info(
                            f"[Step {idx+1}] Tool: {name} | Input: {input_json}"
                        )
                        result = tool.execute(input_json)
                        logging.info(f"[Tool:{name}] Result: {result}")
                        results.append(result)
                        history.append(
                            f"Tool `{name}` executed with input {input_json} produced: {result}"
                        )

                    elif step_type == "llm":
                        llm_query = step.get("query")
                        history_str = "\n".join(history)
                        full_prompt = f"""
Give the answer for the {llm_query} based on the following.

The user asked for the action

{action}

We subdivided the action based to these steps:

{steps}

Here is what has happened so far in the execution:

{history_str}

Use values from history when possible.
"""
                        logging.info(f"[Step {idx+1}] LLM Query: {llm_query}")
                        response = self.llm.get_response(full_prompt.strip())
                        logging.info(f"[LLM] Response: {response}")
                        results.append(response)
                        history.append(f"LLM responded with: {response}")

                    else:
                        msg = f"Unknown step type: {step_type}"
                        logging.error(msg)
                        results.append(msg)
                        history.append(msg)
                history_str = "\n".join(history)
                summary_prompt = f"""
Give the answer without giving the full history. Just answer based on the action in brief.
Give the desired answer only or tell whether the desired outcome happened or not.

The user asked for the action

{action}

We subdivided the action based to these steps:

{steps}

Here is what has happened so far in the execution:

{history_str}

Now, Give the final answer in brief based on the action desired without going into detail of the history.
"""
                final_result = self.llm.get_response(summary_prompt.strip())
                logging.info(f"[Final Summary] {final_result}")
                return final_result
            except Exception as e:
                attempt += 1
                error = e
                logging.warning(f"Retry {attempt} due to error: {e}")



import json

from agents_for_all.agent import Agent
from agents_for_all.llms.base_model import Model
from agents_for_all.tools.python import Python


class DummyLLM(Model):
    def __init__(self):
        self.call_count = 0

    def get_response(self, query: str) -> str:
        if self.call_count == 0:
            self.call_count += 1
            return json.dumps(
                [
                    {
                        "type": "tool",
                        "name": "Python",
                        "input_json": {"code": "print(1 + 1)"},
                    }
                ]
            )
        else:
            return "2"  # Final summary


def test_agent_runs_tool():
    llm = DummyLLM()
    python_tool = Python()
    agent = Agent(llm=llm, tools=[python_tool])
    result = agent.do("Add 1 and 1")
    assert "2" in result



from agents_for_all.llms.direct import Direct


def test_direct_initialization():
    model = Direct(api_endpoint="http://localhost:1234", model="test-model")
    assert model.api_endpoint == "http://localhost:1234"
    assert model.model == "test-model"



from abc import ABC, abstractmethod
from typing import Dict


class Tool(ABC):
    """
    Abstract base class for tools that can be used by agents.

    Tools perform tasks or operations when invoked with structured input.
    Each subclass must implement the `execute` method and define a `description`
    for the LLM or agent to understand the capability of the tool.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the tool.
        Should help the Agent run the appropriate tool.

        Returns:
            str: Name of the tool.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def description(self) -> str:
        """
        A human-readable description of what the tool does.
        Should help the LLM choose the appropriate tool.

        Returns:
            str: Description of the tool's purpose and capabilities.
        """
        raise NotImplementedError

    @abstractmethod
    def execute(self, input_json: Dict) -> str:
        """
        Execute the tool's functionality based on input JSON.

        Args:
            input_json (Dict): Input parameters to guide tool behavior.

        Returns:
            str: Output string describing the result of execution.
        """
        raise NotImplementedError



import logging
import re
from typing import Dict

import requests

from agents_for_all.llms.base_model import Model


class Direct(Model):
    """
    Direct model (or model connector to be exact) which connects to an LLM
    by using api_endpoint and parameters using OpenAI format (as done by LLMStudio).
    """

    def __init__(
        self, api_endpoint: str, model: str, parameters: Dict | None = None
    ) -> None:
        """
        Initialize the Direct model.

        Args:
            api_endpoint (str): The api endpoint to call to get the response.
            model (str): The name of model.
            parameters: (Dict, optional): Other parameters to be used while getting responses. Optional.

        Returns:
            None
        """
        self.api_endpoint = api_endpoint
        self.model = model
        self.parameters = parameters

    def get_response(self, query: str):
        """
        Get response from the LLM based on the given query.

        Args:
            query (str): The query the LLM should respond to.

        Returns:
            str: The response to the query from the LLM.
        """
        json = {
            "messages": [{"role": "user", "content": query}],
            "model": self.model,
        }
        if self.parameters:
            json.update(self.parameters)
        response = requests.post(self.api_endpoint, json=json)
        try:
            response_json = response.json()
            llm_response = response_json["choices"][0]["message"]["content"]
            return re.sub(r"<think>.*?</think>", "", llm_response, flags=re.DOTALL)

        except Exception as e:
            logging.error(e)
